package metaid

//go:generate abigen --sol ./IdentityController.sol --pkg metaid --out ./identity.go

import (
	"crypto/ecdsa"
	"flag"
	"fmt"
	"math/big"
	"os"
	"strings"
	"testing"

	"./lib/test/ens"

	"github.com/ethereum/go-ethereum/accounts/abi"
	"github.com/ethereum/go-ethereum/accounts/abi/bind"
	"github.com/ethereum/go-ethereum/accounts/abi/bind/backends"
	"github.com/ethereum/go-ethereum/common"
	"github.com/ethereum/go-ethereum/core"
	"github.com/ethereum/go-ethereum/crypto"
	"github.com/ethereum/go-ethereum/crypto/sha3"
	"github.com/ethereum/go-ethereum/log"
)

var (
	privkeys  []*ecdsa.PrivateKey
	addresses []common.Address
	keysrc    = []string{
		"foo",
		"bar",
		"xyzzy",
		"plugh",
	}
	hashfunc = sha3.NewKeccak256()
	metaids  = make([][32]byte, 4)
	metadata = make([][32]byte, 4)
	idsrc    = []string{
		"blinky",
		"pinky",
		"inky",
		"clyde",
	}
	debugdebugflag = flag.Bool("vv", false, "veryverbose")
	debugflag      = flag.Bool("v", false, "verbose")
)

func init() {
	flag.Parse()
	loglevel := log.LvlInfo
	if *debugflag {
		loglevel = log.LvlDebug
	} else if *debugdebugflag {
		loglevel = log.LvlTrace
	}
	hs := log.StreamHandler(os.Stderr, log.TerminalFormat(true))
	hf := log.LvlFilterHandler(loglevel, hs)
	h := log.CallerFileHandler(hf)
	log.Root().SetHandler(h)

	for i, src := range keysrc {
		hashfunc.Reset()
		hashfunc.Write([]byte(src))
		key := crypto.ToECDSA(hashfunc.Sum(nil))
		privkeys = append(privkeys, key)
		addresses = append(addresses, crypto.PubkeyToAddress(key.PublicKey))
		log.Info("Generate key", "#", i, "src", src, "privkey", fmt.Sprintf("%x", crypto.FromECDSA(privkeys[i])), "address", fmt.Sprintf("%x", addresses[i]))

	}

	for i, src := range idsrc {
		hashfunc.Reset()
		hashfunc.Write([]byte(src))
		copy(metaids[i][:], hashfunc.Sum(nil))
		hashfunc.Write([]byte("meta"))
		copy(metadata[i][:], hashfunc.Sum(nil))
		log.Info("Generate metaid", "#", i, "src", src, "metaid", fmt.Sprintf("%x", metaids[i]), "metadata", fmt.Sprintf("%x", metadata[i]))
	}
}

func newTestBackend() *backends.SimulatedBackend {
	return backends.NewSimulatedBackend(core.GenesisAlloc{
		addresses[0]: {Balance: big.NewInt(10000000000)},
		addresses[1]: {Balance: big.NewInt(10000000000)},
		addresses[2]: {Balance: big.NewInt(10000000000)},
		addresses[3]: {Balance: big.NewInt(10000000000)},
	})
}

// unit test for signature store
// uses fake bytes32 for simplicity
// IRL we will use ENS node namehashes
func TestSignatureStore(t *testing.T) {
	backend := newTestBackend()
	deployTransactor := bind.NewKeyedTransactor(privkeys[0])
	deployTransactor.Value = big.NewInt(0)

	// deploy contract
	signaturestore_addr, signaturestore, err := deploySignatureStore(privkeys[0], big.NewInt(0), backend)
	if err != nil {
		fmt.Println("err", err)
	}
	backend.Commit()
	log.Debug("Deploy SignatureStore", "address", signaturestore_addr)

	// register one signature and check match
	metadataitem := "aux/foo"
	var metadataitemid [32]byte
	hashfunc.Reset()
	hashfunc.Write([]byte(metadataitem))
	copy(metadataitemid[:], hashfunc.Sum(nil))
	_, err = signaturestore.SetSignature(deployTransactor, metaids[1], metaids[2], metadataitemid, metadata[1], metadata[2])
	if err != nil {
		t.Fatal(err)
	}
	backend.Commit()

	signaturecount, _ := signaturestore.GetSignatureCount(&bind.CallOpts{}, metaids[2], metadataitemid)
	if signaturecount.Cmp(big.NewInt(1)) != 0 {
		t.Fatalf("Signature count mismtach for %x/%x: expected 1, got %v", metaids[2], metadataitem, signaturecount)
	}

	signermetaid, metadatahash, signaturehash, _ := signaturestore.GetSignatureByIndex(&bind.CallOpts{}, metaids[2], metadataitemid, 0)
	if signaturehash != metadata[2] {
		t.Fatalf("Signature mismatch for %x/%x: expected %x, got %x", metaids[2], metadataitem, metadata[2], signaturehash)
	}
	if metadatahash != metadata[1] {
		t.Fatalf("Metadata mismatch for %x/%x: expected %x, got %x", metaids[2], metadataitem, metadata[1], metadatahash)
	}
	if signermetaid != metaids[1] {
		t.Fatalf("Signer metaid mismatch for %x/%x: expected %x, got %x", metaids[2], metadataitem, metaids[1], signermetaid)
	}

	// change store owner
	_, err = signaturestore.SetOwner(deployTransactor, addresses[1])
	if err != nil {
		t.Fatal(err)
	}
	backend.Commit()

	// adding signature should now fail (there should still be only 1 after add)
	_, err = signaturestore.SetSignature(deployTransactor, metaids[0], metaids[2], metadataitemid, metadata[0], metadata[3])
	if err != nil {
		t.Fatal(err)
	}
	backend.Commit()

	signaturecount, _ = signaturestore.GetSignatureCount(&bind.CallOpts{}, metaids[2], metadataitemid)
	if signaturecount.Cmp(big.NewInt(1)) != 0 {
		t.Fatalf("Signature count mismtach for %x/%x: expected 1, got %v", metaids[2], metadataitem, signaturecount)
	}
}

func TestIdentityRegistrar(t *testing.T) {
	backend := newTestBackend()
	deployTransactor := bind.NewKeyedTransactor(privkeys[0])
	deployTransactor.Value = big.NewInt(0)

	// prepare ens node namehashes and hashes
	ensmetanamehash := ens.Namehash(hashfunc, [32]byte{}, []byte("meta"))
	ensidmetanamehash := ens.Namehash(hashfunc, ensmetanamehash, []byte("id"))
	ensmetahash := [32]byte{}
	ensidhash := [32]byte{}
	hashfunc.Reset()
	hashfunc.Write([]byte("id"))
	copy(ensidhash[:], hashfunc.Sum(nil))
	hashfunc.Reset()
	hashfunc.Write([]byte("meta"))
	copy(ensmetahash[:], hashfunc.Sum(nil))

	// deploy all necessary contracts
	ensregistry_addr, ensregistry, err := deployENS(privkeys[0], big.NewInt(0), backend)
	if err != nil {
		t.Fatal(err)
	}
	identityregistrar_addr, identityregistrar, err := deployIdentityRegistrar(privkeys[0], big.NewInt(0), backend, ensregistry_addr, ensidmetanamehash)
	if err != nil {
		t.Fatal(err)
	}
	backend.Commit()

	// set "id.meta" domain node ownership to the registrar
	_, err = ensregistry.SetSubnodeOwner(deployTransactor, [32]byte{}, ensmetahash, addresses[0])
	if err != nil {
		t.Fatal(err)
	}
	backend.Commit()
	_, err = ensregistry.SetSubnodeOwner(deployTransactor, ensmetanamehash, ensidhash, identityregistrar_addr)
	if err != nil {
		t.Fatal(err)
	}
	backend.Commit()

	// register a subdomain
	_, err = identityregistrar.Register(deployTransactor, metaids[0], addresses[1])
	if err != nil {
		t.Fatal(err)
	}
	backend.Commit()

	// check the single hashed node
	namehash := ens.Namehash(hashfunc, ensidmetanamehash, []byte(idsrc[0]))
	metaidowner, err := ensregistry.Owner(&bind.CallOpts{}, namehash)
	if err != nil {
		t.Fatal(err)
	}
	if metaidowner != addresses[1] {
		t.Fatalf("Subdomain owner mismatch: Expected %x, got %x", addresses[1], metaidowner)
	}

	// re-register subdomain
	// (should fail)
	_, err = identityregistrar.Register(deployTransactor, metaids[0], addresses[2])
	if err != nil {
		t.Fatal(err)
	}
	backend.Commit()

	// check that owner still same
	metaidowner, err = ensregistry.Owner(&bind.CallOpts{}, namehash)
	if err != nil {
		t.Fatal(err)
	}
	if metaidowner != addresses[1] {
		t.Logf("Subdomain owner mismatch: Expected %x, got %x", addresses[1], metaidowner)
	}
}

// - create registry, registrar and resolver
// - create identity contract
// - transfer ownership of "id.meta" subdomain to registrar
// - create subdomain through identity contract
// - set content on new subdomain with resolver
func TestIdentity(t *testing.T) {
	backend := newTestBackend()
	deployTransactor := bind.NewKeyedTransactor(privkeys[0])
	deployTransactor.Value = big.NewInt(0)

	// prepare ens node namehashes and hashes
	ensmetanamehash := ens.Namehash(hashfunc, [32]byte{}, []byte("meta"))
	ensidmetanamehash := ens.Namehash(hashfunc, ensmetanamehash, []byte("id"))
	ensmetahash := [32]byte{}
	ensidhash := [32]byte{}
	hashfunc.Reset()
	hashfunc.Write([]byte("id"))
	copy(ensidhash[:], hashfunc.Sum(nil))
	hashfunc.Reset()
	hashfunc.Write([]byte("meta"))
	copy(ensmetahash[:], hashfunc.Sum(nil))

	// deploy all necessary contracts
	ensregistry_addr, ensregistry, err := deployENS(privkeys[0], big.NewInt(0), backend)
	if err != nil {
		t.Fatal(err)
	}
	identityresolver_addr, identityresolver, err := deployIdentityResolver(privkeys[0], big.NewInt(0), backend, ensregistry_addr)
	if err != nil {
		t.Fatal(err)
	}
	identityregistrar_addr, identityregistrar, err := deployIdentityRegistrar(privkeys[0], big.NewInt(0), backend, ensregistry_addr, ensidmetanamehash)
	if err != nil {
		t.Fatal(err)
	}
	signaturestore_addr, signaturestore, err := deploySignatureStore(privkeys[0], big.NewInt(0), backend)
	if err != nil {
		t.Fatal(err)
	}
	identity_addr, identity, err := deployIdentity(privkeys[0], big.NewInt(0), backend)
	if err != nil {
		t.Fatal(err)
	}
	backend.Commit()

	// identity registry and signature store contracts are now owned by privkeys[0]
	// transfer ownership to identity contract
	_, err = identityregistrar.SetOwner(deployTransactor, identity_addr)
	if err != nil {
		fmt.Println("err", err)
	}
	backend.Commit()

	_, err = signaturestore.SetOwner(deployTransactor, identity_addr)
	if err != nil {
		fmt.Println("err", err)
	}
	backend.Commit()

	_, err = identity.SetStores(deployTransactor, identityregistrar_addr, signaturestore_addr)
	if err != nil {
		fmt.Println("err", err)
	}

	log.Info("Deploy Identity", "address", identity_addr, "registraraddress", identityregistrar_addr, "resolveraddress", identityresolver_addr, "registryaddr", ensregistry_addr)

	// set "id.meta" domain node ownership to the registrar
	_, err = ensregistry.SetSubnodeOwner(deployTransactor, [32]byte{}, ensmetahash, addresses[0])
	if err != nil {
		t.Fatal(err)
	}
	backend.Commit()
	_, err = ensregistry.SetSubnodeOwner(deployTransactor, ensmetanamehash, ensidhash, identityregistrar_addr)
	if err != nil {
		t.Fatal(err)
	}
	backend.Commit()

	// register a subdomain and check
	label := common.ToHex(metaids[0][:8])[2:]
	enslabelnamehash := ens.Namehash(hashfunc, ensidmetanamehash, []byte(label))
	enslabelhash := [32]byte{}
	hashfunc.Reset()
	hashfunc.Write([]byte(label))
	copy(enslabelhash[:], hashfunc.Sum(nil))
	_, err = identity.Register(deployTransactor, enslabelhash, addresses[0])
	if err != nil {
		t.Fatal(err)
	}
	backend.Commit()

	metaidowner, err := ensregistry.Owner(&bind.CallOpts{}, enslabelnamehash)
	if err != nil {
		t.Fatal(err)
	}
	if metaidowner != addresses[0] {
		t.Fatalf("Subdomain owner mismatch: Expected %x, got %x", addresses[0], metaidowner)
	}

	// set content hash for the subdomain and check
	_, err = identityresolver.SetContent(deployTransactor, enslabelnamehash, metadata[0])
	if err != nil {
		t.Fatal(err)
	}
	backend.Commit()

	content, err := identityresolver.Content(&bind.CallOpts{}, enslabelnamehash)
	if err != nil {

	}
	if content != metadata[0] {
		t.Fatalf("Content for node %x does not match: expected %x, got %x", metadata[0], content)
	}
}

func TestIdentityController(t *testing.T) {
	backend := newTestBackend()
	deployTransactor := bind.NewKeyedTransactor(privkeys[0])
	deployTransactor.Value = big.NewInt(0)

	// deploy registry contract
	ensregistry_addr, ensregistry, err := deployENS(privkeys[0], big.NewInt(0), backend)
	if err != nil {
		t.Fatal(err)
	}
	backend.Commit()

	// deploy identity controller contract
	var controllerkey [32]byte
	hashfunc.Reset()
	hashfunc.Write([]byte("one"))
	copy(controllerkey[:], hashfunc.Sum(nil))

	identitycontroller_addr, identitycontroller, err := deployIdentityController(privkeys[0], big.NewInt(0), backend)
	if err != nil {
		t.Fatal(err)
	}

	// create a new identity contract with identitycontroller
	_, err = identitycontroller.CreateIdentity(deployTransactor, ensregistry_addr)
	if err != nil {
		t.Fatal(err)
	}
	backend.Commit()

	// get the created contract address
	identity_addr, err := identitycontroller.Current(&bind.CallOpts{})
	if err != nil {
		t.Fatal(err)
	}

	// hack a contract instance
	parsedabi, err := abi.JSON(strings.NewReader(IdentityABI))
	if err != nil {
		t.Fatal(err)
	}
	log.Info("Identity contract found", "addr", fmt.Sprintf("%x", identity_addr))
	identitycontract := bind.NewBoundContract(identity_addr, parsedabi, backend, backend)
	identity := &Identity{
		IdentityCaller: IdentityCaller{
			contract: identitycontract,
		},
		IdentityTransactor: IdentityTransactor{
			contract: identitycontract,
		},
	}

	// retrieve the registrar address of the new identity contract
	identityregistrar_addr, err := identity.IdentityRegistrar(&bind.CallOpts{})
	if err != nil {
		t.Fatal(err)
	}

	// retrieve the signatures store address of the new identity contract
	identitysignaturestore_addr, err := identity.SignatureStore(&bind.CallOpts{})
	if err != nil {
		t.Fatal(err)
	}

	// set ownership of the relevant registry path to the newly created identity's registry
	ensmetanamehash := ens.Namehash(hashfunc, [32]byte{}, []byte("meta"))
	ensidmetanamehash := ens.Namehash(hashfunc, ensmetanamehash, []byte("id"))
	ensmetahash := [32]byte{}
	ensidhash := [32]byte{}
	hashfunc.Reset()
	hashfunc.Write([]byte("id"))
	copy(ensidhash[:], hashfunc.Sum(nil))
	hashfunc.Reset()
	hashfunc.Write([]byte("meta"))
	copy(ensmetahash[:], hashfunc.Sum(nil))

	_, err = ensregistry.SetSubnodeOwner(deployTransactor, [32]byte{}, ensmetahash, addresses[0])
	if err != nil {
		t.Fatal(err)
	}
	backend.Commit()
	_, err = ensregistry.SetSubnodeOwner(deployTransactor, ensmetanamehash, ensidhash, identityregistrar_addr)
	if err != nil {
		t.Fatal(err)
	}
	backend.Commit()

	// check that ownership matches
	metaidowner, err := ensregistry.Owner(&bind.CallOpts{}, ensidmetanamehash)
	if err != nil {
		t.Fatal(err)
	}
	if identityregistrar_addr != metaidowner {
		t.Fatalf("Wrong registry owner: Expected %x, got %x", identityregistrar_addr, metaidowner)
	}

	// create an external identity contract to upgrade to
	newidentity_addr, newidentity, err := deployIdentity(privkeys[0], big.NewInt(0), backend)
	if err != nil {
		t.Fatal(err)
	}
	backend.Commit()

	// set identitycontroller as owner of identity
	_, err = newidentity.SetOwner(deployTransactor, identitycontroller_addr)
	if err != nil {
		t.Fatal(err)
	}
	backend.Commit()

	// upgrade identity contract
	hashfunc.Write([]byte("two"))
	copy(controllerkey[:], hashfunc.Sum(nil))
	_, err = identitycontroller.UpgradeIdentity(deployTransactor, newidentity_addr)
	if err != nil {
		t.Fatal(err)
	}
	backend.Commit()

	// get the upgraded identity contract address
	identityupgrade_addr, err := identitycontroller.Current(&bind.CallOpts{})
	if err != nil {
		t.Fatal(err)
	}
	if identityupgrade_addr == identity_addr {
		t.Fatalf("Identity contract address unchanged: old %x, new %x", identity_addr, identityupgrade_addr)
	}
	log.Info("Upgraded identity contract found", "addr", fmt.Sprintf("%x", identityupgrade_addr))

	identityupgradecontract := bind.NewBoundContract(identityupgrade_addr, parsedabi, backend, backend)
	identityupgrade := &Identity{
		IdentityCaller: IdentityCaller{
			contract: identityupgradecontract,
		},
		IdentityTransactor: IdentityTransactor{
			contract: identityupgradecontract,
		},
	}

	// check if we still have the same registry
	identityupgraderegistrar_addr, err := identityupgrade.IdentityRegistrar(&bind.CallOpts{})
	if err != nil {
		t.Fatal(err)
	}
	if identityupgraderegistrar_addr != identityregistrar_addr {
		t.Fatalf("Identity owner mismatch: expected %x, got %x", identityregistrar_addr, identityupgraderegistrar_addr)
	}

	// check if we still have the same signature store
	identityupgradesignaturestore_addr, err := identityupgrade.SignatureStore(&bind.CallOpts{})
	if err != nil {
		t.Fatal(err)
	}
	if identityupgradesignaturestore_addr != identitysignaturestore_addr {
		t.Fatalf("Identity owner mismatch: expected %x, got %x", identitysignaturestore_addr, identityupgradesignaturestore_addr)
	}
}

func deployENS(prvKey *ecdsa.PrivateKey, amount *big.Int, backend *backends.SimulatedBackend) (common.Address, *ens.ENS, error) {
	deployTransactor := bind.NewKeyedTransactor(prvKey)
	deployTransactor.Value = amount

	addr, _, obj, err := ens.DeployENS(deployTransactor, backend)
	if err != nil {
		return common.Address{}, nil, err
	}
	backend.Commit()
	return addr, obj, nil
}

func deployIdentityResolver(prvKey *ecdsa.PrivateKey, amount *big.Int, backend *backends.SimulatedBackend, ensaddr common.Address) (common.Address, *IdentityResolver, error) {
	deployTransactor := bind.NewKeyedTransactor(prvKey)
	deployTransactor.Value = amount

	addr, _, obj, err := DeployIdentityResolver(deployTransactor, backend, ensaddr)
	if err != nil {
		return common.Address{}, nil, err
	}
	backend.Commit()
	return addr, obj, nil
}

func deployIdentity(prvKey *ecdsa.PrivateKey, amount *big.Int, backend *backends.SimulatedBackend) (common.Address, *Identity, error) {
	deployTransactor := bind.NewKeyedTransactor(prvKey)
	deployTransactor.Value = amount

	addr, _, obj, err := DeployIdentity(deployTransactor, backend)
	if err != nil {
		return common.Address{}, nil, err
	}
	backend.Commit()
	return addr, obj, nil
}

func deployIdentityRegistrar(prvKey *ecdsa.PrivateKey, amount *big.Int, backend *backends.SimulatedBackend, ensaddress common.Address, rootnamehash [32]byte) (common.Address, *IdentityRegistrar, error) {
	deployTransactor := bind.NewKeyedTransactor(prvKey)
	deployTransactor.Value = amount

	addr, _, obj, err := DeployIdentityRegistrar(deployTransactor, backend, ensaddress, rootnamehash)
	if err != nil {
		return common.Address{}, nil, err
	}
	backend.Commit()
	return addr, obj, nil
}

func deploySignatureStore(prvKey *ecdsa.PrivateKey, amount *big.Int, backend *backends.SimulatedBackend) (common.Address, *SignatureStore, error) {
	deployTransactor := bind.NewKeyedTransactor(prvKey)
	deployTransactor.Value = amount

	addr, _, obj, err := DeploySignatureStore(deployTransactor, backend)
	if err != nil {
		return common.Address{}, nil, err
	}
	backend.Commit()
	return addr, obj, nil
}

func deployIdentityController(prvKey *ecdsa.PrivateKey, amount *big.Int, backend *backends.SimulatedBackend) (common.Address, *IdentityController, error) {
	deployTransactor := bind.NewKeyedTransactor(prvKey)
	deployTransactor.Value = amount

	addr, _, obj, err := DeployIdentityController(deployTransactor, backend)
	if err != nil {
		return common.Address{}, nil, err
	}
	backend.Commit()
	return addr, obj, nil
}
