# `setMetadataSignature(label, address)`

Sign a META Identity metadata item.

## Parameters

- `signerMetaId` `<String>` - ENS namehash of metadata signer's META-ID
- `targetMetaId` `<String>` - ENS namehash of metadata owner's META-ID
- `metadataKeyHash` `<String>` - SHA3 hash of signed metadata item key
- `metadataValueHash` `<String>` - SHA3 hash of signed metadata item value
- `signatureContentHash` `<String>` - Content hash of detached signature

## Returns

[Transaction object](http://web3js.readthedocs.io/en/1.0/web3-eth-contract.html#id12)
for the `setMetadataSignature` method, which can then be encoded, signed and sent.

## Example

```js
import MetaId from 'meta-id'
import Web3 from 'web3'
// import META Identity contract addresses
import contractAddresses from '../contracts/addresses.json'

// init meta-id
const metaId = MetaId({ addresses: contractAddresses })
// init web3.js
const web3 = new Web3('http://localhost:8545')

// Ethereum account to send transactions from
const account = web3.eth.accounts.create()

const setMetadataSignatureTransaction = () => {
  // create transaction object
  const tx = metaId.setMetadataSignature(
    metaId.generateMetaId('alice'), // use META-ID = `alice.id.meta`
    metaId.generateMetaId('bob'), // use META-ID = `bob.id.meta`
    web3.utils.sha3('aux/twitter'), // sign 'twitter' metadata item
    web3.utils.sha3('@bob'), // value of 'twitter' metadata item
    `0x...` // content address of signature, eg. Swarm hash
  )

  // encode contract method ABI
  const abi = tx.encodeABI()

  // sign and send transaction
  return account.signTransaction({
    data: abi,
    gas: 4700000,
    to: contractAddresses.Identity,
  })
    .then(signedTxData =>
      web3.eth.sendSignedTransaction(signedTxData.rawTransaction)
    )
    .then(txReceipt => console.log(txReceipt))
}
```
