import MetaId from 'meta-id'
import { privateToPublic } from 'ethereumjs-util'
import Web3 from 'web3'

/**
 * Import META Identity contract addresses
 * JSON file generated from contract deployment scripts
 * @see https://github.com/meta-network/meta-id/blob/master/doc/deploy-contracts.md
 * @type {Object}
 */
import contractAddresses from '../contracts/addresses.json'

/**
 * RPC host
 * @type {String}
 */
const RPC_PROVIDER = 'http://localhost:8545'

/**
 * Swarm host
 * @example - Local node:  http://localhost:8500
 * @example - Web gateway: http://swarm-gateways.net
 * @type {String}
 */
const SWARM_PROVIDER = 'http://localhost:8500'

/**
 * Instantiate Web3 with RPC provider
 * @type {Web3}
 */
const web3 = new Web3(new Web3.providers.HttpProvider(RPC_PROVIDER))

/**
 * Require Web3 `accounts` package
 * @see http://web3js.readthedocs.io/en/1.0/web3-eth-accounts.html
 */
const Accounts = web3.eth.accounts

/**
 * Require Web3 `bzz` package
 * @see http://web3js.readthedocs.io/en/1.0/web3-bzz.html
 */
const Bzz = web3.bzz

/**
 * Set Swarm provider
 * @see http://web3js.readthedocs.io/en/1.0/web3-bzz.html#setprovider
 */
Bzz.setProvider(SWARM_PROVIDER)

/**
 * Instantiate MetaId with contract addresses
 * @type {Object}
 */
const metaId = MetaId({ addresses: contractAddresses })

/**
 * Log the meta-id library version
 */
console.log('🆔 Running meta-id version:', metaId.version())

/**
 * Temporary global state
 * @type {Object}
 */
const state = {}

/**
 * Global state getter
 * @param  {String} key State property key
 * @return              State property value at key
 */
const getState = key => {
  const value = state[key]

  if (!value) {
    throw new Error(`No value for ${key} found in state.`)
  }

  return value
}

/**
 * Swarm URL constructor utility
 *
 * @param  {String} hash Swarm hash
 * @return {String}      Swarm URL for given hash
 */
const getSwarmURL = hash => {
  const formattedHash = hash.substring(0, 2) === '0x' ? hash.slice(2) : hash

  return `${Bzz.currentProvider}/bzz:/${formattedHash}/?list=true`
}

/**
 * Swarm Upload utility method
 *
 * Formats data for upload into structure required by `web3.bzz`
 * Uploads data to Swarm
 * Returns resulting Swarm hash
 *
 * @see https://github.com/MaiaVictor/swarm-js/#uploads
 * @see http://web3js.readthedocs.io/en/1.0/web3-bzz.html#id7
 *
 * @param  {Object} data Data to upload to Swarm
 * @return {String}      Swarm hash
 */
const bzzUpload = data => {
  return Bzz.upload(
    Object.keys(data)
      .map(key => ({
        [`/${key}`]: {
          type: 'text/plain',
          data: JSON.stringify(data[key]),
        },
      }))
      .reduce((o, v) => Object.assign({}, o, v), {})
  )
}

/**
 * Swarm upload event handler
 *
 * @param  {String} swarmhash Swarm hash
 */
const onSwarmUploadSuccess = swarmhash => {
  // log resources
  swarmResourceLogger(swarmhash)

  // store Swarm hash for later use
  return (state.swarmHash = swarmhash)
}

/**
 * Utility method to log Swarm upload URLs
 *
 * @param  {String} hash Swarm hash
 * @return {String}      Log message
 */
const swarmResourceLogger = hash => {
  console.log(`
    🐝🐝🐝
    Uploaded directory. Address: ${hash}
    Manifest: ${SWARM_PROVIDER}/bzzr:/${hash}?content_type=text/plain
    Index: ${SWARM_PROVIDER}/bzz:/${hash}/?list=true
    Content /meta: ${SWARM_PROVIDER}/bzz:/${hash}/meta
    Content /aux: ${SWARM_PROVIDER}/bzz:/${hash}/aux
    Content /key: ${SWARM_PROVIDER}/bzz:/${hash}/key
    🐝🐝🐝
  `)
}

/**
 * Handle register identity event
 */
const onRegisterIdentity = () => {
  // common name is used to generate the META-ID
  const commonName = document.querySelector('.commonName').value
  // Spotity ID and Twitter handle are used as simple examples
  // IRL this could be a wide range of metadata items
  const spotifyId = document.querySelector('.spotifyId').value
  const twitterHandle = document.querySelector('.twitterHandle').value

  // use current connected Ethereum account to send transactions
  const account = getState('currentAccount')

  /**
   * Upload data to Swarm
   * Log resulting Swarm URLs
   * Register new identity
   * Set identity's metadata address
   */
  return bzzUpload({
    meta: {
      id: metaId.generateMetaId(commonName),
      name: commonName,
      parent: '0x',
    },
    aux: {
      ecdsa: privateToPublic(account.privateKey),
      spotifyId: spotifyId,
      twitter: twitterHandle,
    },
  })
    .then(swarmhash => onSwarmUploadSuccess(swarmhash))
    .then(() => registerIdentityTransaction(account, commonName))
    .then(() =>
      setMetadataContentTransaction(account, commonName, state.swarmHash)
    )
    .catch(e => console.error(e))
}

/**
 * Handle retrieve identity event
 */
const onRetrieveIdentity = () => {
  // get identity name to retrieve
  const metaIdName = document.querySelector('.metaIdName').value

  // use current connected Ethereum account to send transaction
  const account = getState('currentAccount')

  // call resolveIdentity transaction to retrieve metadata address
  return resolveIdentityTransaction(account, metaIdName)
    .then(metadataAddress => getSwarmURL(metadataAddress))
    .then(metadataURL =>
      document.querySelector('.metadataURL').setAttribute('href', metadataURL)
    )
}

/**
 * Handle sign identity event
 */
const onSignIdentity = () => {
  // get user submitted data
  const signerMetaIdName = document.querySelector('.signerMetaIdName').value
  const targetMetaIdName = document.querySelector('.targetMetaIdName').value
  const metadataKey = document.querySelector('.metadataKey').value
  const metadataValue = document.querySelector('.metadataValue').value

  // use current connected Ethereum account to send transactions
  const account = getState('currentAccount')

  // sign metadata item value with Ethereum account keys
  const signature = account.sign(metadataValue)

  /**
   * Upload signature to Swarm
   * Register signature against META Identity
   */
  return bzzUpload({ signature }).then(signatureContentHash =>
    setMetadataSignatureTransaction(
      account,
      metadataKey,
      metadataValue,
      signatureContentHash,
      signerMetaIdName,
      targetMetaIdName
    )
  )
}

/**
 * Handle verify identity event
 */
const onVerifyIdentity = () => {
  // get user submitted data
  const verifyMetaIdName = document.querySelector('.verifyMetaIdName').value
  const verifyMetadataKey = document.querySelector('.verifyMetadataKey').value

  // use current connected Ethereum account to send transactions
  const account = getState('currentAccount')

  // generate META-ID from target identity's name
  const targetMetaId = metaId.generateMetaId(verifyMetaIdName)

  // generate SHA3 of metadata key being verified
  const metadataKeyHash = web3.utils.sha3(verifyMetadataKey)

  /**
   * Get number of signatures for metadata item
   * Get each signature content hash
   * Fetch each encoded signature from Swarm
   * Verify each signature against signer's Ethereum account
   */
  return getMetadataSignatureCountTransaction(
    account,
    metadataKeyHash,
    targetMetaId
  )
    .then(signatureCount =>
      Promise.all(
        Array(Number(signatureCount))
          .fill()
          .map((x, signatureIndex) =>
            getMetadataSignatureByIndexTransaction(
              account,
              metadataKeyHash,
              targetMetaId,
              signatureIndex
            )
          )
      )
    )
    .then(metadataSignatures =>
      Promise.all(
        metadataSignatures.map(metadataSignature =>
          Bzz.download(metadataSignature[2].slice(2))
        )
      )
    )
    .then(encodedSignatures =>
      encodedSignatures.map(encodedSignature =>
        JSON.parse(encodedSignature.signature.data.toString())
      )
    )
    .then(signatures =>
      signatures.map(
        signature => Accounts.recover(signature) === account.address
      )
    )
    .then(verificationResults =>
      verificationResults.every(verification => verification)
    )
    .then(isMetadataItemVerified =>
      console.log(`Metadata item verified? ${isMetadataItemVerified}`)
    )
}

/**
 * Handle create account event
 */
const onCreateAccount = () => {
  // get password for encrypting keystore file
  const password = document.querySelector('.generatePassword').value

  // create new Ethereum account
  const account = Accounts.create()

  // create encrypted keystore object
  const encryptedKeystore = Accounts.encrypt(account.privateKey, password)

  // save keystore in memory ready for download
  const blob = new Blob([JSON.stringify(encryptedKeystore)], {
    type: 'application/json',
  })
  const url = URL.createObjectURL(blob)

  // generate keystore filename using meta-id utility
  const filename = metaId.generateKeystoreFilename(encryptedKeystore.address)

  // prepare download button
  const downloadButton = document.querySelector('.downloadKeystore')
  downloadButton.setAttribute('href', url)
  downloadButton.setAttribute('download', filename)

  // render public address
  document.querySelector('.ethereumAddress').textContent = account.address

  // set connected account to newly created account
  state.currentAccount = account
}

/**
 * Handle upload account event
 */
const onUploadAccount = e => {
  const reader = new FileReader()
  const file = e.target.files[0]

  reader.onload = e => {
    // parse the uploaded file into a JSON object
    const encryptedKeystoreString = e.target.result
    const encryptedKeystoreObject = JSON.parse(encryptedKeystoreString)

    // get password for decrypting keystore file
    const password = document.querySelector('.uploadPassword').value

    // decrypt keystore object
    const decryptedKeystore = Accounts.decrypt(
      encryptedKeystoreObject,
      password
    )

    // render public address
    document.querySelector('.ethereumAddress').textContent =
      decryptedKeystore.address

    // set connected account to uploaded account
    state.currentAccount = decryptedKeystore
  }

  // invoke file reader
  reader.readAsText(file)
}

/**
 * Handle local Ether faucet request event
 */
const onCreditAccount = () => {
  // get currently connected account
  const account = getState('currentAccount')

  // send funds from coinbase to connected account
  return web3.eth.getAccounts().then(accounts => accounts[4]).then(coinbase =>
    web3.eth.sendTransaction({
      from: coinbase,
      gas: 4700000,
      to: account.address,
      value: 999999999999999999,
    })
  )
}

/**
 * Handle IdentityRegistered contract event
 *
 * @see http://web3js.readthedocs.io/en/1.0/web3-eth-contract.html#contract-events-return
 *
 * @param  {Object} event Contract event object
 */
const onIdentityRegistered = event => {
  console.info(`
    🆕 New META Identity registered:
    ENS node: ${event.returnValues.node}
    Owner address: ${event.returnValues.newowner}
  `)
}

/**
 * Handle IdentityUpdated contract event
 *
 * @see http://web3js.readthedocs.io/en/1.0/web3-eth-contract.html#contract-events-return
 *
 * @param  {Object} event Contract event object
 */
const onIdentityUpdated = event => {
  console.info(`
    🆙 META Identity metadata updated:
    ENS node: ${event.returnValues.node}
    Content hash: ${event.returnValues.hash}
  `)
}

/**
 * Handle IdentitySigned contract event
 *
 * @see http://web3js.readthedocs.io/en/1.0/web3-eth-contract.html#contract-events-return
 *
 * @param  {Object} event Contract event object
 */
const onIdentitySigned = event => {
  console.info(`
    ✍️ META Identity metadata item signed:
    META-ID: ${event.returnValues.target}
    Metadata item key hash: ${event.returnValues.item_id_digest}
    Signature content hash: ${event.returnValues.signature_digest}
  `)
}

/**
 * Create, sign and send registerIdentity transaction
 *
 * @param  {String} address    Ethereum account address
 * @param  {String} commonName Name of identity owner
 * @return {Object}            Transaction receipt
 */
const registerIdentityTransaction = (account, commonName) => {
  // create transaction object
  const tx = metaId.registerIdentity(
    web3.utils.sha3(commonName),
    account.address
  )

  // encode contract method ABI
  const abi = tx.encodeABI()

  // sign and send transaction
  return account
    .signTransaction({
      data: abi,
      gas: 4700000,
      to: contractAddresses.Identity,
    })
    .then(signedTransactionData =>
      web3.eth.sendSignedTransaction(signedTransactionData.rawTransaction)
    )
}

/**
 * Create, sign and send setMetadataContent transaction
 *
 * @param  {Object} account    Ethereum account
 * @param  {String} commonName Name of identity owner
 * @param  {String} swarmHash  Metadata Swarm hash
 * @return {Object}            Transaction receipt
 */
const setMetadataContentTransaction = (account, commonName, swarmHash) => {
  // create transaction object
  const tx = metaId.setMetadataContent(
    metaId.generateMetaId(commonName),
    `0x${swarmHash}`
  )

  // encode contract method ABI
  const abi = tx.encodeABI()

  // sign and send transaction
  return account
    .signTransaction({
      data: abi,
      gas: 4700000,
      to: contractAddresses.IdentityResolver,
    })
    .then(signedTransactionData =>
      web3.eth.sendSignedTransaction(signedTransactionData.rawTransaction)
    )
}

/**
 * Create and call resolveIdentity transaction
 *
 * @param  {Object} account    Ethereum account
 * @param  {String} metaIdName Name of META Identity
 * @return {Object}            Transaction receipt
 */
const resolveIdentityTransaction = (account, metaIdName) => {
  // create transaction object
  const tx = metaId.resolveIdentity(metaId.generateMetaId(metaIdName))

  // call transaction
  return tx.call({ from: account.address, gas: 4700000 })
}

/**
 * Create, sign and send setMetadataSignature transaction
 *
 * @param  {Object} account              Ethereum account
 * @param  {String} metadataKey          Metadata item key
 * @param  {String} metadataValue        Metadata item value
 * @param  {String} signatureContentHash Signature content hash
 * @param  {String} signerMetaIdName     Signer META Identity name
 * @param  {String} targetMetaIdName     Target META Identity name
 * @return {Object}                      Transaction receipt
 */
const setMetadataSignatureTransaction = (
  account,
  metadataKey,
  metadataValue,
  signatureContentHash,
  signerMetaIdName,
  targetMetaIdName
) => {
  // create transaction object
  const tx = metaId.setMetadataSignature(
    metaId.generateMetaId(signerMetaIdName),
    metaId.generateMetaId(targetMetaIdName),
    web3.utils.sha3(metadataKey),
    web3.utils.sha3(metadataValue),
    `0x${signatureContentHash}`
  )

  // encode contract method ABI
  const abi = tx.encodeABI()

  // sign and send transaction
  return account
    .signTransaction({
      data: abi,
      gas: 4700000,
      to: contractAddresses.Identity,
    })
    .then(signedTransactionData =>
      web3.eth.sendSignedTransaction(signedTransactionData.rawTransaction)
    )
}

/**
 * Create and call getMetadataSignatureCount transaction
 *
 * @param  {Object} account         Ethereum account
 * @param  {String} metadataKeyHash Hash of metadata item key
 * @param  {String} targetMetaId    Target META-ID
 * @return {Object}                 Transaction receipt
 */
const getMetadataSignatureCountTransaction = (
  account,
  metadataKeyHash,
  targetMetaId
) => {
  // create transaction object
  const tx = metaId.getMetadataSignatureCount(targetMetaId, metadataKeyHash)

  // call transaction
  return tx.call({ from: account.address, gas: 4700000 })
}

/**
 * Create and call getMetadataSignatureByIndex transaction
 *
 * @param  {Object} account         Ethereum account
 * @param  {String} metadataKeyHash Hash of metadata item key
 * @param  {String} targetMetaId    Target META-ID
 * @param  {String} signatureIndex  Index of signature in target signatures array
 * @return {Object}                 Transaction receipt
 */
const getMetadataSignatureByIndexTransaction = (
  account,
  metadataKeyHash,
  targetMetaId,
  signatureIndex
) => {
  // create transaction object
  const tx = metaId.getMetadataSignatureByIndex(
    targetMetaId,
    metadataKeyHash,
    signatureIndex
  )

  // call transaction
  return tx.call({ from: account.address, gas: 4700000 })
}

/**
 * Contract event subscriptions
 */
console.log(metaId.events)
metaId.events.IdentityRegistered().on('data', onIdentityRegistered)
metaId.events.IdentityUpdated().on('data', onIdentityUpdated)
metaId.events.IdentitySigned().on('data', onIdentitySigned)

/**
 * Event Listeners
 */
document
  .querySelector('.registerIdentity')
  .addEventListener('click', onRegisterIdentity)
document
  .querySelector('.retrieveIdentity')
  .addEventListener('click', onRetrieveIdentity)
document
  .querySelector('.signIdentity')
  .addEventListener('click', onSignIdentity)
document
  .querySelector('.verifyIdentity')
  .addEventListener('click', onVerifyIdentity)
document
  .querySelector('.createAccount')
  .addEventListener('click', onCreateAccount)
document
  .querySelector('.uploadAccount')
  .addEventListener('change', onUploadAccount)
document
  .querySelector('.creditAccount')
  .addEventListener('click', onCreditAccount)
