const args = require('yargs').argv
const chalk = require('chalk')
const fs = require('fs-extra')
const namehash = require('eth-ens-namehash')
const path = require('path')
const prettier = require('prettier')
const Web3 = require('web3')

/**
 * Options for Prettier code formatter
 *
 * @todo - these should be in a shared config
 *
 * @type {Object}
 */
const prettierOptions = {
  semi: false,
  singleQuote: true,
  trailingComma: 'es5',
}

/**
 * Import ENS Registry Contract ABI and bytecode
 */
const ensInterface = require('./utils/ens')

/**
 * Default JSON-RPC connection endpoint
 * @type {String}
 */
const DEFAULT_RPC = 'http://localhost:8545'

/**
 * ENS subnode label for META Network
 * @type {String}
 */
const META_SUBNODE_LABEL = 'meta'

/**
 * ENS subnode label for META Identity
 * @type {String}
 */
const META_ID_SUBNODE_LABEL = 'id'

/**
 * Get CLI args
 * @const
 */
const { config, rpc } = args

/**
 * Import config options
 * @const opts
 */
const opts = require(path.resolve(config))

/**
 * Resolve deployed contract addresses filepath
 * @type {String}
 */
const deployedAddressesPath = path.resolve(opts.addressesPath)

/**
 * Import contract interfaces
 * @const compiledContracts
 */
const compiledContracts = require(path.resolve(opts.contractsPath))

/**
 * Initialise Web3
 * @type {Web3}
 */
const web3 = new Web3(
  new Web3.providers.HttpProvider(opts.rpc[rpc] || DEFAULT_RPC)
)

/**
 * Create the Identity contracts
 *
 * @todo - this script is crazy long and in need of abstraction/refactor
 *
 * 1. Deploy the IdentityController contract
 * 2. Create the Identity abstract contract from IdentityController
 * 3. Deploy the Identity contract
 * 4. Set IdentityController contract as the Identity contract owner
 * 5. Upgrade the current Identity contract from IdentityController
 * 6. Create META Identity ENS subnode with IdentityRegistrar contract as owner
 * 7. Deploy the IdentityResolver contract
 * 8. Export Identity & IdentityResolver contract interfaces to `/lib`
 *
 * @example node create-identity.js --config ./relative.path.to.config.js
 */
const createIdentity = (async () => {
  try {
    // contract deployment address not specified?
    if (!opts.fromAccount) {
      // get web3 accounts
      const accounts = await web3.eth.getAccounts()

      // set contract deployment address to primary account
      opts.fromAccount = accounts[0]
    }

    // ensure deployed contract addresses file exists before reading it
    const deployedAddressesExists = await fs.pathExists(deployedAddressesPath)

    // create empty JSON file if deployed contract addresses file does not exist
    if (!deployedAddressesExists)
      await fs.outputJson(deployedAddressesPath, {}, { spaces: 2 })

    // read existing deployed contract addresses
    const deployedAddresses = await fs.readJson(deployedAddressesPath)

    // throw error if ENS contract not already deployed
    if (!deployedAddresses.ENS) {
      throw new Error('Run `npm run deploy-ens` before running this script')
    }

    // get Identity, IdentityController and IdentityResolver contract interfaces
    const identityInterface = compiledContracts.Identity
    const identityControllerInterface = compiledContracts.IdentityController
    const identityResolverInterface = compiledContracts.IdentityResolver

    // create IdentityController contract instance
    const identityControllerContract = new web3.eth
      .Contract(identityControllerInterface.abi, {
      data: identityControllerInterface.bytecode,
    })

    // deploy IdentityController contract
    const identityController = await identityControllerContract
      .deploy()
      .send({ from: opts.fromAccount, gasLimit: opts.gasLimit })

    // get IdentityController contract address
    const identityControllerAddress =
      identityController.contractAddress || identityController.options.address

    // set IdentityController contract address
    identityControllerContract.options.address = identityControllerAddress

    // send IdentityController.createIdentity method
    const createIdentityTx = await identityControllerContract.methods
      .createIdentity(deployedAddresses.ENS)
      .send({ from: opts.fromAccount, gasLimit: opts.gasLimit })

    // create Identity contract instance
    const identityContract = new web3.eth.Contract(identityInterface.abi, {
      data: identityInterface.bytecode,
    })

    // deploy Identity contract
    const identity = await identityContract
      .deploy()
      .send({ from: opts.fromAccount, gasLimit: opts.gasLimit })

    // get Identity contract address
    const identityAddress = identity.contractAddress || identity.options.address

    // set Identity contract instance address
    identityContract.options.address = identityAddress

    // send Identity.setOwner method
    const setOwnerTx = await identityContract.methods
      .setOwner(identityControllerAddress)
      .send({ from: opts.fromAccount, gasLimit: opts.gasLimit })

    // send IdentityController.upgradeIdentity method
    const upgradeIdentityTx = await identityControllerContract.methods
      .upgradeIdentity(identityAddress)
      .send({ from: opts.fromAccount, gasLimit: opts.gasLimit })

    // get upgraded Identity contract address from upgradeIdentity tx event
    const identityUpgradedAddress =
      upgradeIdentityTx.events.IdentityUpgraded.returnValues.upgradedidentity

    // update Identity contract instance with upgraded Identity contract address
    identityContract.options.address = identityUpgradedAddress

    // call Identity.identityRegistrar method to get IdentityRegistrar contract address
    const identityRegistrarAddress = await identityContract.methods
      .identityRegistrar()
      .call({ from: opts.fromAccount, gasLimit: opts.gasLimit })

    // create ENS contract instance at deployed address
    const ensContract = new web3.eth.Contract(
      ensInterface.abi,
      deployedAddresses.ENS
    )

    // send ENS.setSubnodeOwner method to create `id` subnode of `meta` subnode [id.meta]
    // and give ownership to deployed IdentityRegistrar contract
    const setSubnodeOwnerTx = await ensContract.methods
      .setSubnodeOwner(
        namehash.hash(META_SUBNODE_LABEL),
        web3.utils.sha3(META_ID_SUBNODE_LABEL),
        identityRegistrarAddress
      )
      .send({ from: opts.fromAccount, gasLimit: opts.gasLimit })

    // create IdentityResolver contract instance
    const identityResolverContract = new web3.eth
      .Contract(identityResolverInterface.abi, {
      data: identityResolverInterface.bytecode,
    })

    // deploy IdentityResolver contract
    const identityResolver = await identityResolverContract
      .deploy({ arguments: [deployedAddresses.ENS] })
      .send({ from: opts.fromAccount, gasLimit: opts.gasLimit })

    // get IdentityResolver contract address
    const identityResolverAddress =
      identityResolver.contractAddress || identityResolver.options.address

    // write deployed contract addresses to disk
    await fs.writeJson(
      deployedAddressesPath,
      Object.assign({}, deployedAddresses, {
        Identity: identityUpgradedAddress,
        IdentityController: identityControllerAddress,
        IdentityRegistrar: identityRegistrarAddress,
        IdentityResolver: identityResolverAddress,
      }),
      { spaces: 2 }
    )

    // construct Identity module export
    const identityContractExport = prettier.format(
      `export const abi = ${JSON.stringify(identityInterface.abi)}

      export const Identity = new web3.eth.Contract(abi)

      export const setAddress = address => Identity.options.address = address
    `,
      prettierOptions
    )

    // construct IdentityResolver module export
    const identityResolverContractExport = prettier.format(
      `export const abi = ${JSON.stringify(identityResolverInterface.abi)}

      export const IdentityResolver = new web3.eth.Contract(abi)

      export const setAddress = address => IdentityResolver.options.address = address
    `,
      prettierOptions
    )

    // write Identity contract export file
    await fs.writeFileSync(
      `${opts.outputPath}/${opts.contracts.Identity.filename}.js`,
      opts.outputString.concat(identityContractExport)
    )

    // write IdentityResolver contract export file
    await fs.writeFileSync(
      `${opts.outputPath}/${opts.contracts.IdentityResolver.filename}.js`,
      opts.outputString.concat(identityResolverContractExport)
    )

    return console.log(chalk.green.bold(`🎉 Identity contract created!`))
  } catch (e) {
    console.error(e)
  }
})()
