const args = require('yargs').argv
const chalk = require('chalk')
const fs = require('fs-extra')
const path = require('path')
const Web3 = require('web3')

/**
 * Import ENS Registry Contract ABI and bytecode
 */
const { abi, bytecode } = require('./utils/ens')

/**
 * Default JSON-RPC connection endpoint
 * @type {String}
 */
const DEFAULT_RPC = 'http://localhost:8545'

/**
 * ENS subnode label for META Network
 * @type {String}
 */
const META_SUBNODE_LABEL = 'meta'

/**
 * Get CLI args
 * @const
 */
const { config, rpc } = args

/**
 * Import config options
 * @const opts
 */
const opts = require(path.resolve(config))

/**
 * Resolve deployed contract addresses filepath
 * @type {String}
 */
const deployedAddressesPath = path.resolve(opts.addressesPath)

/**
 * Initialise Web3
 * @type {Web3}
 */
const web3 = new Web3(
  new Web3.providers.HttpProvider(opts.rpc[rpc] || DEFAULT_RPC)
)

/**
 * Deploy the ENS Registry to the specified RPC node
 * Create and own META Network subnode
 *
 * @see https://docs.ens.domains/en/latest/deploying.html
 *
 * @example node deploy-ens.js --config ./relative.path.to.config.js
 */
const deployENS = (async () => {
  // contract deployment address not specified?
  if (!opts.fromAccount) {
    // get web3 accounts
    const accounts = await web3.eth.getAccounts()

    // set contract deployment address to primary account
    opts.fromAccount = accounts[0]
  }

  // create ENS contract instance
  const ensContract = new web3.eth.Contract(abi, { data: bytecode })

  // deploy ENS contract
  const ens = await ensContract
    .deploy()
    .send({ from: opts.fromAccount, gasLimit: opts.gasLimit })

  // get deployed ENS contract address
  const ensAddress = ens.contractAddress || ens.options.address

  // set ENS contract address
  ensContract.options.address = ensAddress

  // send ENS.setSubnodeOwner method to create and own `meta` subnode of rootNode `0`
  const setSubnodeOwnerTx = await ensContract.methods
    .setSubnodeOwner(
      '0x0',
      web3.utils.sha3(META_SUBNODE_LABEL),
      opts.fromAccount
    )
    .send({ from: opts.fromAccount, gasLimit: opts.gasLimit })

  // ensure deployed contract addresses file exists before reading it
  const deployedAddressesExists = await fs.pathExists(deployedAddressesPath)

  // create empty JSON file if deployed contract addresses file does not exist
  if (!deployedAddressesExists)
    await fs.outputJson(deployedAddressesPath, {}, { spaces: 2 })

  // read existing deployed contract addresses
  const deployedAddresses = await fs.readJson(deployedAddressesPath)

  // write deployed contract address to disk
  await fs.writeJson(
    deployedAddressesPath,
    Object.assign({}, deployedAddresses, {
      ENS: ensAddress,
    }),
    { spaces: 2 }
  )

  return console.log(chalk.green.bold(`🌐 ENS deployed!`))
})()
