const args = require('yargs').argv
const chalk = require('chalk')
const fs = require('fs-extra')
const path = require('path')
const prettier = require('prettier')
const Web3 = require('web3')

/**
 * Options for Prettier code formatter
 *
 * @todo - these should be in a shared config
 *
 * @type {Object}
 */
const prettierOptions = {
  semi: false,
  singleQuote: true,
  trailingComma: 'es5',
}

/**
 * Default JSON-RPC connection endpoint
 * @type {String}
 */
const DEFAULT_RPC = 'http://localhost:8545'

/**
 * Get CLI args
 * @const
 */
const { config, rpc } = args

/**
 * Import config options
 * @const opts
 */
const opts = require(path.resolve(config))

/**
 * Import contract addresses
 * @const deployedAddresses
 */
const deployedAddresses = require(path.resolve(opts.addressesPath))

/**
 * Import contract interfaces
 * @const compiledContracts
 */
const compiledContracts = require(path.resolve(opts.contractsPath))

/**
 * Initialise Web3
 * @type {Web3}
 */
const web3 = new Web3(
  new Web3.providers.HttpProvider(opts.rpc[rpc] || DEFAULT_RPC)
)

/**
 * Upgrade the Identity contract
 *
 * 1. Deploy Identity contract
 * 2. Set IdentityController contract as the Identity contract owner
 * 3. Upgrade to latest Identity contact from IdentityController contract
 *
 * @example node upgrade-identity.js --config ./relative.path.to.config.js
 */
const upgradeIdentity = (async () => {
  try {
    // contract deployment address not specified?
    if (!opts.fromAccount) {
      // get web3 accounts
      const accounts = await web3.eth.getAccounts()

      // set contract deployment address to primary account
      opts.fromAccount = accounts[0]
    }

    // get Identity and IdentityController contract interfaces
    const identityInterface = compiledContracts.Identity
    const identityControllerInterface = compiledContracts.IdentityController

    // create Identity contract instance
    const identityContract = new web3.eth.Contract(identityInterface.abi, {
      data: identityInterface.bytecode,
    })

    // deploy Identity contract
    const identity = await identityContract
      .deploy()
      .send({ from: opts.fromAccount, gasLimit: opts.gasLimit })

    // get Identity contract address
    const identityAddress = identity.contractAddress || identity.options.address

    // set Identity contract instance address
    identityContract.options.address = identityAddress

    // create IdentityController contract instance at deployed address
    const identityControllerContract = new web3.eth.Contract(
      identityControllerInterface.abi,
      deployedAddresses.IdentityController
    )

    // send Identity.setOwner method
    const setOwnerTx = await identityContract.methods
      .setOwner(deployedAddresses.IdentityController)
      .send({ from: opts.fromAccount, gasLimit: opts.gasLimit })

    // send IdentityController.upgradeIdentity method
    const upgradeIdentityTx = await identityControllerContract.methods
      .upgradeIdentity(identityAddress)
      .send({ from: opts.fromAccount, gasLimit: opts.gasLimit })

    // get upgraded Identity contract address from upgradeIdentity tx event
    const identityUpgradedAddress =
      upgradeIdentityTx.events.IdentityUpgraded.returnValues.upgradedidentity

    // write deployed contract addresses to disk
    await fs.writeJson(
      opts.addressesPath,
      Object.assign({}, deployedAddresses, {
        Identity: identityUpgradedAddress,
      }),
      { spaces: 2 }
    )

    // construct Identity module export
    const identityContractExport = prettier.format(
      `export const abi = ${JSON.stringify(identityInterface.abi)}

      export const Identity = new web3.eth.Contract(abi)

      export const setAddress = address => Identity.options.address = address
    `,
      prettierOptions
    )

    // write Identity contract export file
    await fs.writeFileSync(
      `${opts.outputPath}/${opts.contracts.Identity.filename}.js`,
      opts.outputString.concat(identityContractExport)
    )

    return console.log(chalk.green.bold(`🎉 Identity contract upgraded!`))
  } catch (e) {
    console.error(e)
  }
})()
